// Cornell University Department of Electrical Engineering
// Master of Engineering (Electrical) Design Project
// Wei-jiunn Jang
// AVRStudio / ATmega644

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

//set up the debugging utility ASSERT
#define __ASSERT_USE_STDERR
#include <assert.h>
#include "uart.h"
// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);


unsigned char enter_pressed;


//timeout values for each task
#define t1 100  
#define t2 100


//menu state machine state names
#define main_menu 1 
#define choose_sensor 2
#define set_threshold 21
#define control_pump 22
#define release_pump 23
#define celsius 4
#define fahrenheit 5



#define begin {
#define end   }

//the  subroutines
void task1(void);
void task2(void);
void getstr_int(void);	// Starts getting a string from serial line
void putstr_int(void);	// Starts a send to serial line
void initialize(void); 	// All the usual mcu stuff 
          
volatile int time1, time2;		// task scheduling timeout counters
char v[10];
char newstr ;

// when code added is invalid, set flag;
unsigned char invalid;



// for loop index
unsigned char i;
	
// RXC ISR variables								
volatile unsigned char r_index;  //current string index
volatile char r_buffer[16];	//input string 
volatile char r_ready;  //flag for receive done
volatile char r_char;  //current character  

// TX ISR variables								
volatile unsigned char t_index;  //current string index
volatile char t_buffer[300];	//output string 
volatile char t_ready;  //flag for transmit done
volatile char t_char;  //current character


// toggling LED variable
unsigned char LED;

// variables for main fuctions
unsigned char display_menu;
unsigned char MENU;



//display TIME and MENU subroutines
void show_menu(void);

// bit handling macros for i/o registers
#define READ(U, N) ((U) >> (N) & 1u)
#define SET(U, N) ((void)((U) |= 1u << (N)))
#define CLR(U, N) ((void)((U) &= ~(1u << (N))))
#define FLIP(U, N) ((void)((U) ^= 1u << (N)))

char Hbits, Lbits;
double temp[2];
int threshold;
char debug_led;
char status = 0;
char valve[4];
char force_status[4];

char force = 0;


// Celsius mode(0) or Fahrenheit(1)
char degree = 0;

#define set_temp 1
#define control 2
#define release 3





    
//**********************************************************
//timer 0 overflow ISR
ISR (TIMER0_COMPA_vect)
begin
	//Decrement the times if they are not already zero
	if (time1>0)	--time1;
	if (time2>0) 	--time2;
end  

//**********************************************************
// UART character-ready ISR
// builds a sting and signals when the string is complete
// supports backspace

ISR (USART0_RX_vect)
begin
	r_char = UDR0 ;    //get a char

	//build the input string
	if (r_char != '\r') 	// Is the input a <enter>?
	begin 
		if (r_char == '\b')	// Is the input a backspace? 
		begin
			if (r_index > 0) 
			begin
				putchar('\b');	// backup
				putchar(' '); 	// erase the character on the screen
				putchar('\b');	// backup
				--r_index ;		// wipe a character from the string
			end
		end
		else if (r_char == 27)
		begin
			MENU = main_menu;
			display_menu = 0;
			putchar(12);
		end
		else if (r_char == ' ');	// ignore all the space
		else   
		begin
			UDR0 = r_char;
			r_buffer[r_index++] = r_char ; // add a character to the string
		end
	end
	else 					// Human pressed <enter>
	begin
		putchar('\n');  			//use putchar to avoid overwrite
		r_buffer[r_index] = 0x00;   //zero terminate
		r_ready = 1;				//signal cmd processor
		UCSR0B ^= (1<<RXCIE0) ;   	//stop rec ISR -- clear rxc
	end
end

/**********************************************************/
//UART xmit-empty ISR
ISR (USART0_UDRE_vect)
begin
	t_char = t_buffer[++t_index];
	if (t_char == 0) 			// end of string?	
	begin
		UCSR0B ^= (1<<UDRIE0) ; 	// kill isr -- clear tx enable
		t_ready = 1; 			// transmit done 
	end
	else	UDR0 = t_char ;  	//send the char 
end

//**********************************************************       
//Entry point and task scheduler loop
int main(void)
begin  
	initialize();

	while(1)
	begin     
		if (time1==0){time1=t1;	task1(); }
		if (time2==0){time2=t2;	task2(); }
	end   
end  
  


//**********************************************************          
//Task 1   input a string and print it
void task1(void) 
begin 
  //print ad get another serial string      
  if (r_ready)    
  begin
  	enter_pressed = 1;
	// read string which is now ready
	sscanf(r_buffer,"%s",v);
	if (r_buffer[0] == 0) v[0] = 0;
	
	// set up procedure to get the next string input
	// and read it using receive ISR
 	getstr_int();
  end
end  
 
//********************************************************** 
//Task 2  print the meaasge and system time in seconds
void task2(void) 
begin

	// update sensor value
	debug_led = debug_led ^ 1;	// blink LED on the prototype board (PORTD.2)
	PORTD = (debug_led<<2);

	for(int i=0;i<2;i++)
	begin
		CLR(PORTB,i+3);		//chip select low begins transmission
		// send signal to get High byte
		SPDR = 0;
		//wait until SPIF is set (SPI done)
		while (!(SPSR & (1<<SPIF)));
		Hbits = SPDR;		// get back Hbits
		
		// send signal to get Low byte
		SPDR = 0;
		//wait until SPIF is set (SPI done)
		while (!(SPSR & (1<<SPIF)));
		Lbits = SPDR;		// get back Lbits
		SET(PORTB,i+3);		//chip select high ends transmission

		// calculate the temperature
		temp[i] = (double)Hbits * 2;
		temp[i] = temp[i] + ((double)(Lbits >> 5))/4;


		if (force == 0)		// only change status when there's no force on/off
		begin
			if (temp[0] >= threshold)
			begin
				status = 1;
				valve[0] = 'O';
				valve[1] = 'n';
				valve[2] = ' ';
			end
			else if (temp[0] < temp[1]+2)
			begin
				status = 0;
				valve[0] = 'O';
				valve[1] = 'f';
				valve[2] = 'f';
			end
		end

		if (status == 1)
			PORTC = 0x01;
		else
			PORTC = 0;

	end


	switch (MENU)
	begin

	case main_menu:
		if (enter_pressed == 0)		// when enter is not pressed
		begin
			if (display_menu == 0)
			begin
				putchar(12);
				show_menu();
				display_menu = 1;
			end
			MENU = main_menu;
		end
		else						// when enter is pressed
		begin
			if (strlen(v) == 1)
			begin						// check what was entered

				if (v[0] == '1')
				begin
					if (!degree) fprintf(stdout,"Please enter threshold temperature (30 C - 99 C) >>");
					else fprintf(stdout,"Please enter threshold temperature (86 F - 210 F) >>");
					MENU = set_threshold;
					enter_pressed = 0;
				end

				else if (v[0] == '2')
				begin
					fprintf(stdout,"Turn On/Off pump? (1/0) >>");
					MENU = control_pump;
					enter_pressed = 0;
				end

				else if (v[0] == '3')
				begin
					fprintf(stdout,"Release control over pump? (y/n) >>");
					MENU = release_pump;
					enter_pressed = 0;
				end

				else if (v[0] == '4')		// change to celsius mode
				begin
					degree = 0;
					MENU = main_menu;
					display_menu = 0;
					enter_pressed = 0;
				end

				else if (v[0] == '5')		// change to fahrenheit mode
				begin
					degree = 1;
					MENU = main_menu;
					display_menu = 0;
					enter_pressed = 0;
				end
				else
				begin
					fprintf(stdout, "invalid entry.\n\r");
					enter_pressed = 0;
				end
			end
			else
			begin
				fprintf(stdout, "invalid entry.\n\r");
				enter_pressed = 0;
			end
		end
		break;


	case set_threshold:
		if (enter_pressed == 1)
		begin
			// when addind a code, first check whether the length is 4
			if (strlen(v) == 2 || strlen(v) == 3)
			begin
				if (!degree)
				begin
					if (atoi(v) >= 30 && atoi(v) <= 99)
					begin
						for(i=0;i<strlen(v);i++)
						begin			// then check whether they are all numbers
							if(v[i]<48 || v[i]>57 ) invalid = 1;
							else invalid = 0;
						end
					end
					else invalid = 1;
				end
				else
				begin
					if (atoi(v) >= 86 && atoi(v) <= 210)
					begin
						for(i=0;i<strlen(v);i++)
						begin			// then check whether they are all numbers
							if(v[i]<48 || v[i]>57 ) invalid = 1;
							else invalid = 0;
						end
					end
					else invalid = 1;
				end

				if (invalid == 0)
				begin
					if (!degree) threshold = atoi(v);
					else threshold = ((atoi(v))- 32) * 5 / 9;
					
					enter_pressed = 0;
					display_menu = 0;
					MENU = main_menu;
				end
				else 
				begin
					invalid = 0;
					if (!degree) fprintf(stdout,"invalid entry.\n\rPlease enter threshold temperature (30 C - 99 C) >>");
					else fprintf(stdout,"invalid entry.\n\rPlease enter threshold temperature (86 F - 210 F) >>");
				end
			end
			else
			begin
				if (!degree) fprintf(stdout,"invalid entry.\n\rPlease enter threshold temperature (30 C - 99 C) >>");
				else fprintf(stdout,"invalid entry.\n\rPlease enter threshold temperature (86 F - 210 F) >>");
			end
			enter_pressed = 0;
		end
		else
			MENU = set_threshold;
		break;

	case control_pump:
		if (enter_pressed == 1)
		begin
			// when choosing sensor, should be 1 or 2
			if (strlen(v) == 1)
			begin
				if (v[0]==48)	//off
				begin
					MENU = main_menu;
					display_menu = 0;
					status = 0;
					valve[0] = 'O';
					valve[1] = 'f';
					valve[2] = 'f';
					force = 1;
					force_status[0] = 'Y';
					force_status[1] = 'e';
					force_status[2] = 's';
				end
				else if (v[0]==49)	//on
				begin
					MENU = main_menu;
					display_menu = 0;
					status = 1;
					valve[0] = 'O';
					valve[1] = 'n';
					valve[2] = ' ';
					force = 1;
					force_status[0] = 'Y';
					force_status[1] = 'e';
					force_status[2] = 's';
				end
				else
				fprintf(stdout, "invalid entry\n\rTurn On/Off Pump? (1/0)");
			end
			else
				fprintf(stdout, "invalid entry\n\rTurn On/Off Pump? (1/0)");
			enter_pressed = 0;
		end
		else
			MENU = control_pump;
		break;
		
	case release_pump:
		if (enter_pressed == 1)
		begin
			// when choosing sensor, should be 1 or 2
			if (strlen(v) == 1)
			begin
				if (v[0] == 121)	//yes
				begin
					MENU = main_menu;
					display_menu = 0;
					force = 0;
					force_status[0] = ' ';
					force_status[1] = ' ';
					force_status[2] = ' ';
				end
				else if (v[0] == 110)	//on
				begin
					MENU = main_menu;
					display_menu = 0;
				end
				else
				fprintf(stdout, "invalid entry\n\rRelease control over pump? (y/n)\n\r>>");
			end
			else
				fprintf(stdout, "invalid entry\n\rRelease control over pump? (y/n)\n\r>>");

			enter_pressed = 0;
		end
		else
			MENU = release_pump;
		break;
	end
end  


  
//********************************************************** 
// Non-blocking keyboard input: initializes ISR-driven receive.
// This routine merely sets up the ISR, which then
//does all the work of getting a command.
void getstr_int(void) 
begin
	r_ready=0;	// mark string as not ready
	r_index=0;	// reset index
	// turn on receive ISR
	UCSR0B |= (1<<RXCIE0) ;
end

//********************************************************** 
// Nonblocking print: initializes ISR-driven transmit. 
// This routine merely sets up the ISR, then
// sends one character, The ISR does all the work.   
void putstr_int(void) 
begin
	t_ready=0;	// mark transmitter as busy
	t_index=0;	// reset index
	// see if there is actually a string
	if (t_buffer[0]>0) 
	begin
		// if so, send the first chararcter
  		putchar(t_buffer[0]);
		// and turn on transmit (UDR empty) ISR
  		UCSR0B |= (1<<UDRIE0) ;
	end 
end


void show_menu(void)
begin
	fprintf(stdout, "------ Cornell University  Solar Decathlon ------\n\r");
	fprintf(stdout, "--- Solar Hot Water Collecting Control System ---\n\r");
	fprintf(stdout, "-------------------------------------------------\n\r");
	fprintf(stdout, "--| Sensor      |-|    Coil    |-|  WaterTank |--\n\r");

		if (!degree)
	fprintf(stdout, "--| Temperature |-| %7.2f C  |-| %7.2f C  |--\n\r",temp[0],temp[1]);
		else
	fprintf(stdout, "--| Temperature |-| %7.2f F  |-| %7.2f F  |--\n\r",temp[0]*1.8+32,temp[1]*1.8+32);
		if (!degree)
	fprintf(stdout, "--| Threshold   |-| %4d    C  |-| %7.2f C  |--\n\r",threshold,temp[1]+2);
		else
			begin
			if( ((double)threshold*1.8+32) - (char)(threshold*1.8+32) > 0)
	fprintf(stdout, "--| Threshold   |-| %4d    F  |-| %7.2f F  |--\n\r",(char)(threshold*1.8+32)+1,(temp[1]+2)*1.8+32);
			else
	fprintf(stdout, "--| Threshold   |-| %4d    F  |-| %7.2f F  |--\n\r",(char)(threshold*1.8+32),(temp[1]+2)*1.8+32);
			end

	fprintf(stdout, "--| Pump Status |-|     %3s    |-|\n\r",valve);
	fprintf(stdout, "--| Controlled? |-|     %3s    |-|\n\r",force_status);
	fprintf(stdout, "-------------------------------------------------\n\r");


	//*******************************************************************************
	//* remember to setup floating point printf properties in project configuration	*
	//* in Libraries																*
	//* add libprintf_flt.a and libm.a												*
	//*																				*
	//* in Custom Options															*
	//* go to [Linker Options] and add												*
	//* -Wl,-u,vfprintf																*
	//*******************************************************************************


	sprintf(t_buffer, "1. Set threshold temperature\n\r2. Manual control \n\r3. Release control\n\r4. Switch to Celsius\n\r5. Switch to Fahrenheit\n\r-------------------------------------------------\n\r>>");

	newstr = 1;

	if (t_ready && newstr)
	begin
		// make a string of system time (sec) and number from task 1
		//sprintf(t_buffer,"%ld %d\n\r",time/1000, v) ;

		// print it using transmit ISR 
		putstr_int();
		// clear the string ready handshake from task 1 
		newstr = 0;
	end
end




//********************************************************** 
//Set it all up
void initialize(void)
begin

	// init the UART -- uart_init() is in uart.c
	uart_init();
	stdout = stdin = stderr = &uart_str;
	fprintf(stdout,"Starting ISR UART demo...\n\r"); 
	putchar(12);

	// set up timer 0     
	OCR0A = 249;	   			// clear after 250 counts
	TIMSK0 = (1<<OCIE0A) ;	//turn on timer 0 cmp-match ISR 

	// turn on timer 0 clear on match
	TCCR0A = (1<<WGM01) ;
	// timer 0 prescalar to 64 
	TCCR0B = 3 ;	
  
	// init the task timers
	time1=t1;
	time2=t2;    
  
	// initialize the USRT handshake flags
	r_ready=0; // initially, there is no input ready
	t_ready=1; // initially, the transmitter is ready

	// initialize main function variables
	display_menu = 0;
	MENU = main_menu;
	enter_pressed = 0;
	invalid = 0;	

	strcpy(force_status, "   ");
	strcpy(valve, "Off");

	// output pin to pump controlling relay
	DDRC = 0x01;
	PORTC = 0;

	DDRD |= (1<<2) ; // LEDs
	debug_led = 0x00;

	// LM70 reading
	//set up i/o data direction
	//DDRB.4 = 1; //output chip select for DAC (notSYNC)
	//DDRB.5 = 1; //output MOSI to ADC
	//DDRB.6 = 0; //input MISO from ADC
	//DDRB.7 = 1; //output SCLK   
	DDRB = (1<<DDB3) | (1<<DDB4) | (1<<DDB5) | (1<<DDB7) ; 

	// initialize notSYNC
	SET(PORTB,4);
	SET(PORTB,3);

	//set up SPI
	//bit 7 SPIE=0 no ISR
	//bit 6 SPE=1 enable spi
	//bit 5 DORD=0 msb first 
	//bit 4 MSTR=1 Mega32 is spi master
	//bit 3 CPOL=1 clock polarity
	//bit 2 CPHA=1 clock phase
	//bit 1,0 rate sel=10 along with SPRC=1 sets clk to f/32 = 500 kHz
	SPCR = (1<<SPE) | (1<<MSTR) | (1<<CPOL) | (1<<CPHA) | (1<<SPR1) ; //SPCR SPCR = 0b01011110 ;                                                                               
	SPSR = (1<<SPI2X) ; //SPSR = 1; 

	// initialize threshold temperature to 50C
	threshold = 50;

	// turn in ISRs
	// but note that UART ISRs are enabled when reading/writing
	sei(); 

	// start waiting for input
	getstr_int();
end  

   
